@@ -44,19 +44,30 @@ def worker_def(A):
                 base_class.write_images_annotations(out_path)
             except:
                 print('failed on {}'.format(base_class.series_instances_dictionary[index]['Image_Path']))
+                fid = open(os.path.join(base_class.series_instances_dictionary[index]['Image_Path'], 'failed.txt'),
+                           'w+')
+                fid.close()
             q.task_done()
 
 
-def sort_xy(x, y):
-    x -= np.min(x)
-    y -= np.min(y)
-    x0 = np.mean(x)
-    y0 = np.mean(y)
-    r = np.sqrt((x - x0) ** 2 + (y - y0) ** 2)
-    angles = np.where((y - y0) > 0, np.arccos((x - x0) / r), 2 * np.pi - np.arccos((x - x0) / r))
-    # mask = np.argsort(angles)
-    # mask = np.abs(mask - np.max(mask))
-    return angles
+def folder_worker(A):
+    q = A[0]
+    while True:
+        item = q.get()
+        if item is None:
+            break
+        else:
+            dicom_adder = AddDicomToDictionary()
+            dicom_path, images_dictionary, rt_dictionary, rd_dictionary = item
+            try:
+                print('Loading from {}'.format(dicom_path))
+                dicom_adder.add_dicom_to_dictionary_from_path(dicom_path=dicom_path,
+                                                              images_dictionary=images_dictionary,
+                                                              rt_dictionary=rt_dictionary,
+                                                              rd_dictionary=rd_dictionary)
+            except:
+                print('failed on {}'.format(dicom_path))
+            q.task_done()
 
 
 class PointOutputMakerClass(object):
@@ -81,27 +92,16 @@ class PointOutputMakerClass(object):
             contours = find_contours(temp_image, level=0.5, fully_connected='low', positive_orientation='high')
             for contour in contours:
                 contour = np.squeeze(contour)
-                slope = (contour[1:, 1] - contour[:-1, 1]) / (contour[1:, 0] - contour[:-1, 0])
+                with np.errstate(divide='ignore'):
+                    slope = (contour[1:, 1] - contour[:-1, 1]) / (contour[1:, 0] - contour[:-1, 0])
                 slope_index = None
                 out_contour = []
                 for index in range(len(slope)):
                     if slope[index] != slope_index:
                         out_contour.append(contour[index])
                     slope_index = slope[index]
-                # angles = sort_xy(x=np.asarray(out_contour)[:, 0], y=np.asarray(out_contour)[:, 1])
                 contour = [[float(c[1]), float(c[0]), float(i)] for c in out_contour]
-                # k = [dicom_handle.TransformPhysicalPointToIndex(zz) for zz in k]
                 contour = np.asarray([dicom_handle.TransformContinuousIndexToPhysicalPoint(zz) for zz in contour])
-                # distances = np.sum((contour[1:] - contour[:-1])**2, axis=1)**.5
-                # spacing = dicom_handle.GetSpacing()
-                # spacing_min = np.min(dicom_handle.GetSpacing())
-                # distance = 0
-                # out_contour = []
-                # for index in range(len(distances)):
-                #     distance += distances[index]
-                #     if distance > spacing_min:
-                #         out_contour.append(contour[index])
-                #         distance = 0
                 self.contour_dict[i].append(np.asarray(contour))
 
 
@@ -112,32 +112,65 @@ def poly2mask(vertex_row_coords, vertex_col_coords, shape):
     return mask
 
 
-def add_images_to_dictionary(series_instances_dictionary, sitk_dicom_reader, path):
+def add_images_to_dictionary(images_dictionary, sitk_dicom_reader, path):
     """
-    :param series_instances_dictionary: dictionary of series instance UIDs
+    :param images_dictionary: dictionary of series instance UIDs for images
     :param sitk_dicom_reader: sitk.ImageFileReader()
     :param path: path to the images or structure in question
     """
     series_instance_uid = sitk_dicom_reader.GetMetaData("0020|000e")
-    patientID = sitk_dicom_reader.GetMetaData("0010|0020")[:-1]
-    description = sitk_dicom_reader.GetMetaData("0008|103e")
-    keys = []
-    series_instance_uids = []
-    for key, value in series_instances_dictionary.items():
-        keys.append(key)
-        series_instance_uids.append(value['SeriesInstanceUID'])
-    temp_dict = {'Image_Path': path, 'Description': description, 'SeriesInstanceUID': series_instance_uid,
-                 'PatientID': patientID}
-    if series_instance_uid not in series_instance_uids:
-        template_dictionary = return_template_dictionary()
-        template_dictionary.update(temp_dict)
-        index = 0
-        while index in series_instances_dictionary:
-            index += 1
-        series_instances_dictionary[index] = template_dictionary
-    else:
-        index = keys[series_instance_uids.index(series_instance_uid)]
-        series_instances_dictionary[index].update(temp_dict)
+    if series_instance_uid not in images_dictionary:
+        patientID = sitk_dicom_reader.GetMetaData("0010|0020")[:-1]
+        description = None
+        if "0008|103e" in sitk_dicom_reader.GetMetaDataKeys():
+            description = sitk_dicom_reader.GetMetaData("0008|103e")
+        study_instance_uid = sitk_dicom_reader.GetMetaData("0020|000d")
+        temp_dict = {'PatientID': patientID, 'SeriesInstanceUID': series_instance_uid,
+                     'StudyInstanceUID': study_instance_uid, 'RTs': {}, 'RDs': {},
+                     'Image_Path': path, 'Description': description}
+        images_dictionary[series_instance_uid] = temp_dict
+
+
+def add_rt_to_dictionary(ds, path, rt_dictionary):
+    """
+    :param ds: pydicom data structure
+    :param path: path to the images or structure in question
+    """
+    series_instance_uid = ds.SeriesInstanceUID
+    if series_instance_uid not in rt_dictionary:
+        for referenced_frame_of_reference in ds.ReferencedFrameOfReferenceSequence:
+            for referred_study_sequence in referenced_frame_of_reference.RTReferencedStudySequence:
+                for referred_series in referred_study_sequence.RTReferencedSeriesSequence:
+                    refed_series_instance_uid = referred_series.SeriesInstanceUID
+                    if Tag((0x3006, 0x020)) in ds.keys():
+                        ROI_Structure = ds.StructureSetROISequence
+                    else:
+                        ROI_Structure = []
+                    rois_in_structure = {}
+                    rois = []
+                    for Structures in ROI_Structure:
+                        rois.append(Structures.ROIName.lower())
+                        if Structures.ROIName not in rois_in_structure:
+                            rois_in_structure[Structures.ROIName] = Structures.ROINumber
+                    temp_dict = {'Path': path, 'ROI_Names': rois, 'ROIs_in_structure': rois_in_structure,
+                                 'SeriesInstanceUID': refed_series_instance_uid}
+                    rt_dictionary[series_instance_uid] = temp_dict
+
+
+def add_rd_to_dictionary(sitk_dicom_reader, rd_dictionary):
+    """
+    :param ds: pydicom data structure
+    :param path: path to the images or structure in question
+    """
+    series_instance_uid = sitk_dicom_reader.GetMetaData("0020|000e")
+    if series_instance_uid not in rd_dictionary:
+        study_instance_uid = sitk_dicom_reader.GetMetaData("0020|000d")
+        description = None
+        if "0008|103e" in sitk_dicom_reader.GetMetaDataKeys():
+            description = sitk_dicom_reader.GetMetaData("0008|103e")
+        temp_dict = {'Path': sitk_dicom_reader.GetFileName(), 'StudyInstanceUID': study_instance_uid,
+                     'Description': description}
+        rd_dictionary[series_instance_uid] = temp_dict
 
 
 def add_sops_to_dictionary(sitk_dicom_reader, series_instances_dictionary):
@@ -159,22 +192,55 @@ def add_sops_to_dictionary(sitk_dicom_reader, series_instances_dictionary):
 
 
 def return_template_dictionary():
-    template_dictionary = {'Image_Path': None, 'PatientID': None, 'RTs': {}, 'Description': None,
+    template_dictionary = {'Image_Path': None, 'PatientID': None, 'RTs': {}, 'RDs': {}, 'Description': None,
                            'SOP_Instance_UIDs': None, 'SeriesInstanceUID': None}
     return template_dictionary
 
 
+class AddDicomToDictionary(object):
+    def __init__(self):
+        self.image_reader = sitk.ImageFileReader()
+        self.image_reader.LoadPrivateTagsOn()
+        self.reader = sitk.ImageSeriesReader()
+        self.reader.GlobalWarningDisplayOff()
+
+    def add_dicom_to_dictionary_from_path(self, dicom_path, images_dictionary, rt_dictionary, rd_dictionary):
+        fileList = []
+        for dirName, dirs, fileList in os.walk(dicom_path):
+            break
+        fileList = [i for i in fileList if i.find('.dcm') != -1]
+        series_ids = self.reader.GetGDCMSeriesIDs(dicom_path)
+        all_names = []
+        for series_id in series_ids:
+            dicom_names = self.reader.GetGDCMSeriesFileNames(dicom_path, series_id)
+            all_names += [os.path.split(i)[-1] for i in dicom_names]
+            self.image_reader.SetFileName(dicom_names[0])
+            self.image_reader.Execute()
+            modality = self.image_reader.GetMetaData("0008|0060")
+            if modality.lower().find('rtdose') != -1:
+                add_rd_to_dictionary(sitk_dicom_reader=self.image_reader,
+                                     rd_dictionary=rd_dictionary)
+            else:
+                add_images_to_dictionary(images_dictionary=images_dictionary,
+                                         sitk_dicom_reader=self.image_reader, path=dicom_path)
+        RT_Files = [os.path.join(dicom_path, file) for file in fileList if file not in all_names]
+        for lstRSFile in RT_Files:
+            rt = pydicom.read_file(lstRSFile)
+            modality = rt.Modality
+            if modality.lower().find('struct') != -1:
+                add_rt_to_dictionary(ds=rt, path=lstRSFile, rt_dictionary=rt_dictionary)
+
+
 class DicomReaderWriter:
-    def __init__(self, description='', rewrite_RT_file=False, delete_previous_rois=True, Contour_Names=None,
-                 verbose=True,
-                 template_dir=None, arg_max=True, create_new_RT=True, require_all_contours=True, associations={},
-                 iteration=0, get_dose_output=False, flip_axes=(False, False, False), index=0,
+    def __init__(self, description='', rewrite_RT_file=False, delete_previous_rois=True, Contour_Names=[],
+                 verbose=True, template_dir=None, arg_max=True, create_new_RT=True, require_all_contours=True,
+                 associations={}, iteration=0, get_dose_output=False, flip_axes=(False, False, False), index=0,
                  series_instances_dictionary={}):
         """
         :param description: string, description information to add to .nii files
         :param rewrite_RT_file: Boolean, should we re-write the RT structure
         :param delete_previous_rois: delete the previous RTs within the structure
-        :param Contour_Names: list of contour nmes
+        :param Contour_Names: list of contour names
         :param template_dir: default to None, specifies path to template RT structure
         :param arg_max: perform argmax on the mask
         :param require_all_contours: Boolean, require all contours present when making nifti files?
@@ -184,32 +250,35 @@ class DicomReaderWriter:
         :param flip_axes: tuple(3), axis that you want to flip, defaults to (False, False, False)
         :param series_instances_dictionary: dictionary of series instance UIDs of images and RTs
         """
+        self.rt_dictionary = {}
+        self.images_dictionary = {}
+        self.rd_dictionary = {}
         self.series_instances_dictionary = series_instances_dictionary
         self.get_dose_output = get_dose_output
         self.require_all_contours = require_all_contours
         self.flip_axes = flip_axes
         self.create_new_RT = create_new_RT
         self.associations = associations
-        self.set_contour_names(Contour_Names)
-        self.__set_associations__(associations)
+        self.Contour_Names = Contour_Names
+        self.set_contour_names_and_associations(Contour_Names=Contour_Names, associations=associations,
+                                                check_contours=False)
         self.__set_description__(description)
         self.__set_iteration__(iteration)
         self.arg_max = arg_max
         self.rewrite_RT_file = rewrite_RT_file
-        self.dose_handles = []
+        self.dose_handle = None
         if template_dir is None or not os.path.exists(template_dir):
             template_dir = os.path.join(os.path.split(__file__)[0], 'template_RS.dcm')
         self.template_dir = template_dir
         self.template = True
         self.delete_previous_rois = delete_previous_rois
         self.reader = sitk.ImageSeriesReader()
-        self.image_reader = sitk.ImageFileReader()
-        self.image_reader.LoadPrivateTagsOn()
         self.reader.MetaDataDictionaryArrayUpdateOn()
         self.reader.LoadPrivateTagsOn()
         self.verbose = verbose
         self.dicom_handle_uid = None
         self.RS_struct_uid = None
+        self.rd_study_instance_uid = None
         self.index = index
         self.all_RTs = {}
         self.RTs_with_ROI_Names = {}
@@ -219,8 +288,63 @@ class DicomReaderWriter:
     def set_index(self, index):
         self.index = index
 
+    def __compile__(self):
+        """
+        The goal of this is to combine image, rt, and dose dictionaries based on the SeriesInstanceUIDs
+        """
+        print('Compiling dictionaries together...')
+        series_instance_uids = []
+        for key, value in self.series_instances_dictionary.items():
+            series_instance_uids.append(value['SeriesInstanceUID'])
+        index = 0
+        image_keys = list(self.images_dictionary.keys())
+        image_keys.sort()
+        for series_instance_uid in image_keys:  # Will help keep things in order later
+            if series_instance_uid not in series_instance_uids:
+                while index in self.series_instances_dictionary:
+                    index += 1
+                self.series_instances_dictionary[index] = self.images_dictionary[series_instance_uid]
+                series_instance_uids.append(series_instance_uid)
+        for rt_series_instance_uid in self.rt_dictionary:
+            series_instance_uid = self.rt_dictionary[rt_series_instance_uid]['SeriesInstanceUID']
+            rt_dictionary = self.rt_dictionary[rt_series_instance_uid]
+            path = rt_dictionary['Path']
+            self.all_RTs[path] = rt_dictionary['ROI_Names']
+            for roi in rt_dictionary['ROI_Names']:
+                if roi not in self.RTs_with_ROI_Names:
+                    self.RTs_with_ROI_Names[roi] = [path]
+                else:
+                    self.RTs_with_ROI_Names[roi].append(path)
+            if series_instance_uid in series_instance_uids:
+                index = series_instance_uids.index(series_instance_uid)
+                self.series_instances_dictionary[index]['RTs'].update({rt_series_instance_uid: self.rt_dictionary[rt_series_instance_uid]})
+            else:
+                while index in self.series_instances_dictionary:
+                    index += 1
+                template = return_template_dictionary()
+                template['RTs'].update({rt_series_instance_uid: self.rt_dictionary[rt_series_instance_uid]})
+                self.series_instances_dictionary[index] = template
+        study_instance_uids = None
+        for rd_series_instance_uid in self.rd_dictionary:
+            if study_instance_uids is None:
+                study_instance_uids = []
+                for key, value in self.series_instances_dictionary.items():
+                    study_instance_uids.append(value['StudyInstanceUID'])
+            study_instance_uid = self.rd_dictionary[rd_series_instance_uid]['StudyInstanceUID']
+            if study_instance_uid in study_instance_uids:
+                index = study_instance_uids.index(study_instance_uid)
+                self.series_instances_dictionary[index]['RDs'].update({rd_series_instance_uid:
+                                                                           self.rd_dictionary[rd_series_instance_uid]})
+            else:
+                while index in self.series_instances_dictionary:
+                    index += 1
+                template = return_template_dictionary()
+                template['RDs'].update({rd_series_instance_uid: self.rd_dictionary[rd_series_instance_uid]})
+                self.series_instances_dictionary[index] = template
+
     def __reset__(self):
         self.__reset_RTs__()
+        self.rd_study_instance_uid = None
         self.dicom_handle_uid = None
         self.series_instances_dictionary = {}
 
@@ -230,6 +354,14 @@ class DicomReaderWriter:
         self.RS_struct_uid = None
         self.RTs_with_ROI_Names = {}
 
+    def set_contour_names_and_associations(self, Contour_Names=None, associations=None, check_contours=True):
+        if Contour_Names is not None:
+            self.__set_contour_names__(Contour_Names=Contour_Names)
+        if associations is not None:
+            self.__set_associations__(associations=associations)
+        if check_contours:  # I don't want to run this on the first build..
+            self.__check_if_all_contours_present__()
+
     def __set_associations__(self, associations={}):
         keys = list(associations.keys())
         for key in keys:
@@ -240,15 +372,13 @@ class DicomReaderWriter:
                     associations[name] = name
         self.associations, self.hierarchy = associations, {}
 
-    def set_contour_names(self, Contour_Names=None):
+    def __set_contour_names__(self, Contour_Names):
         self.__reset_RTs__()
-        if Contour_Names is None:
-            Contour_Names = []
-        else:
-            Contour_Names = [i.lower() for i in Contour_Names]
+        Contour_Names = [i.lower() for i in Contour_Names]
+        for name in Contour_Names:
+            if name not in self.associations:
+                self.associations[name] = name
         self.Contour_Names = Contour_Names
-        self.__set_associations__(self.associations)
-        self.__check_if_all_contours_present__()
 
     def __set_description__(self, description):
         self.desciption = description
@@ -319,22 +449,43 @@ class DicomReaderWriter:
             print('The following indexes have all ROIs present')
             for index in self.indexes_with_contours:
                 print('Index {}, located at {}'.format(index, self.series_instances_dictionary[index]['Image_Path']))
+            return self.indexes_with_contours
         else:
-            print('You need to first define what ROIs you want, please use .set_contour_names(roi_list)')
+            print('You need to first define what ROIs you want, please use'
+                  ' .set_contour_names_and_associations(roi_list)')
 
     def down_folder(self, input_path):
         print('Please move from down_folder() to walk_through_folders()')
         self.walk_through_folders(input_path=input_path)
 
-    def walk_through_folders(self, input_path):
+    def walk_through_folders(self, input_path, thread_count = int(cpu_count() * 0.9 - 1)):
         """
         Iteratively work down paths to find DICOM files, if they are present, add to the series instance UID dictionary
         :param input_path: path to walk
         """
+        paths_with_dicom = []
         for root, dirs, files in os.walk(input_path):
             dicom_files = [i for i in files if i.endswith('.dcm')]
             if dicom_files:
-                self.add_dicom_to_dictionary_from_path(root)
+                paths_with_dicom.append(root)
+                # dicom_adder.add_dicom_to_dictionary_from_path(dicom_path=root, images_dictionary=self.images_dictionary,
+                #                                               rt_dictionary=self.rt_dictionary)
+        if paths_with_dicom:
+            q = Queue(maxsize=thread_count)
+            A = (q,)
+            threads = []
+            for worker in range(thread_count):
+                t = Thread(target=folder_worker, args=(A,))
+                t.start()
+                threads.append(t)
+            for index, path in enumerate(paths_with_dicom):
+                item = [path, self.images_dictionary, self.rt_dictionary, self.rd_dictionary]
+                q.put(item)
+            for i in range(thread_count):
+                q.put(None)
+            for t in threads:
+                t.join()
+            self.__compile__()
         if self.verbose or len(self.series_instances_dictionary) > 1:
             for key in self.series_instances_dictionary:
                 print('Index {}, description {} at {}'.format(key,
@@ -345,89 +496,9 @@ class DicomReaderWriter:
         self.__check_if_all_contours_present__()
         return None
 
-    def add_dicom_to_dictionary_from_path(self, PathDicom):
-        self.PathDicom = PathDicom
-        self.lstFilesDCM = []
-        self.lstRSFile = None
-        self.Dicom_info = []
-        fileList = []
-        self.RTs_in_case = {}
-        self.RDs_in_case = {}
-        for dirName, dirs, fileList in os.walk(PathDicom):
-            break
-        fileList = [i for i in fileList if i.find('.dcm') != -1]
-        series_ids = self.reader.GetGDCMSeriesIDs(self.PathDicom)
-        all_names = []
-        for series_id in series_ids:
-            dicom_names = self.reader.GetGDCMSeriesFileNames(self.PathDicom, series_id)
-            all_names += dicom_names
-            self.reader.SetFileNames(dicom_names)
-            self.image_reader.SetFileName(dicom_names[0])
-            self.image_reader.Execute()
-            add_images_to_dictionary(series_instances_dictionary=self.series_instances_dictionary,
-                                     sitk_dicom_reader=self.image_reader, path=self.PathDicom)
-        RT_Files = [os.path.join(PathDicom, file) for file in fileList if file not in all_names]
-        for lstRSFile in RT_Files:
-            rt = pydicom.read_file(lstRSFile)
-            modality = rt.Modality
-            if modality.lower().find('dose') != -1:
-                self.RDs_in_case[lstRSFile] = []
-            elif modality.lower().find('struct') != -1:
-                self.RTs_in_case[lstRSFile] = []
-                self.add_rt_to_dictionary(ds=rt, path=lstRSFile)
-        # if self.get_images_mask and self.template:
-        #     self.use_template()
-
-    def add_rt_to_dictionary(self, ds, path):
-        """
-        :param ds: pydicom data structure
-        :param path: path to the images or structure in question
-        """
-        series_instance_uid = ds.SeriesInstanceUID
-        keys = []
-        series_instance_uids = []
-        for key, value in self.series_instances_dictionary.items():
-            keys.append(key)
-            series_instance_uids.append(value['SeriesInstanceUID'])
-        for referenced_frame_of_reference in ds.ReferencedFrameOfReferenceSequence:
-            for referred_study_sequence in referenced_frame_of_reference.RTReferencedStudySequence:
-                for referred_series in referred_study_sequence.RTReferencedSeriesSequence:
-                    refed_series_instance_uid = referred_series.SeriesInstanceUID
-                    if refed_series_instance_uid not in series_instance_uids:
-                        index = 0
-                        while index in self.series_instances_dictionary:
-                            index += 1
-                        temp_dict = return_template_dictionary()
-                        temp_dict['SeriesInstanceUID'] = refed_series_instance_uid
-                        temp_dict['PatientID'] = ds.PatientID
-                        self.series_instances_dictionary[index] = temp_dict
-                    else:
-                        index = keys[series_instance_uids.index(refed_series_instance_uid)]
-                    if series_instance_uid not in self.series_instances_dictionary[index]['RTs']:
-                        if Tag((0x3006, 0x020)) in ds.keys():
-                            ROI_Structure = ds.StructureSetROISequence
-                        else:
-                            ROI_Structure = []
-                        rois_in_structure = {}
-                        rois = []
-                        for Structures in ROI_Structure:
-                            rois.append(Structures.ROIName.lower())
-                            if Structures.ROIName not in rois_in_structure:
-                                rois_in_structure[Structures.ROIName] = Structures.ROINumber
-                            if Structures.ROIName.lower() not in self.RTs_with_ROI_Names:
-                                self.RTs_with_ROI_Names[Structures.ROIName.lower()] = [path]
-                            elif path not in self.RTs_with_ROI_Names[Structures.ROIName.lower()]:
-                                self.RTs_with_ROI_Names[Structures.ROIName.lower()].append(path)
-                        temp_dict = {series_instance_uid: {'Path': path, 'ROI_Names': rois,
-                                                           'ROIs_in_structure': rois_in_structure}}
-                        self.all_RTs[path] = rois_in_structure
-                        self.RTs_in_case[path] = rois_in_structure
-                        self.series_instances_dictionary[index]['RTs'].update(temp_dict)
-
     def write_parallel(self, out_path, excel_file, thread_count=int(cpu_count() * 0.9 - 1)):
         if not os.path.exists(out_path):
             os.makedirs(out_path)
-        q = Queue(maxsize=thread_count)
         final_out_dict = {'PatientID': [], 'Path': [], 'Iteration': [], 'Folder': [], 'SeriesInstanceUID': []}
         if os.path.exists(excel_file):
             df = pd.read_excel(excel_file)
@@ -442,12 +513,6 @@ class DicomReaderWriter:
                     'arg_max': self.arg_max, 'require_all_contours': self.require_all_contours,
                     'Contour_Names': self.Contour_Names,
                     'description': self.desciption, 'get_dose_output': True}
-        A = (q,)
-        threads = []
-        for worker in range(thread_count):
-            t = Thread(target=worker_def, args=(A,))
-            t.start()
-            threads.append(t)
         rewrite_excel = False
         '''
         First, build the excel file that we will use to reference iterations, Series UIDs, and paths
@@ -462,7 +527,7 @@ class DicomReaderWriter:
                     iteration += 1
                 temp_dict = {'PatientID': [self.series_instances_dictionary[index]['PatientID']],
                              'Path': [self.series_instances_dictionary[index]['Image_Path']],
-                             'Iteration': [iteration], 'Folder': [None], 'SeriesInstanceUID': [series_instance_uid]}
+                             'Iteration': [int(iteration)], 'Folder': [None], 'SeriesInstanceUID': [series_instance_uid]}
                 temp_df = pd.DataFrame(temp_dict)
                 df = df.append(temp_df)
         if rewrite_excel:
@@ -470,12 +535,19 @@ class DicomReaderWriter:
         '''
         Next, read through the excel sheet and see if the out paths already exist
         '''
+        q = Queue(maxsize=thread_count)
+        A = (q,)
+        threads = []
+        for worker in range(thread_count):
+            t = Thread(target=worker_def, args=(A,))
+            t.start()
+            threads.append(t)
         for index in self.indexes_with_contours:
             series_instance_uid = self.series_instances_dictionary[index]['SeriesInstanceUID']
             previous_run = df.loc[df['SeriesInstanceUID'] == series_instance_uid]
             if previous_run.shape[0] == 0:
                 continue
-            iteration = previous_run['Iteration'].values[0]
+            iteration = int(previous_run['Iteration'].values[0])
             folder = previous_run['Folder'].values[0]
             if pd.isnull(folder):
                 folder = None
@@ -498,6 +570,8 @@ class DicomReaderWriter:
             'Index is not present in the dictionary! Set it using set_index(index)'
         self.get_images()
         self.get_mask()
+        if self.get_dose_output:
+            self.get_dose()
 
     def get_images(self):
         assert self.index in self.series_instances_dictionary, \
@@ -524,12 +598,51 @@ class DicomReaderWriter:
             self.image_size_cols, self.image_size_rows, self.image_size_z = self.dicom_handle.GetSize()
             self.dicom_handle_uid = series_instance_uid
 
+    def get_dose(self):
+        assert self.index in self.series_instances_dictionary, \
+            'Index is not present in the dictionary! Set it using set_index(index)'
+        index = self.index
+        if self.dicom_handle_uid != self.series_instances_dictionary[index]['SeriesInstanceUID']:
+            print('Loading images for index {}, since mask was requested but image loading was '
+                  'previously different\n'.format(index))
+            self.get_images()
+        if self.rd_study_instance_uid == self.series_instances_dictionary[index]['StudyInstanceUID']:  # Already loaded
+            return None
+        self.rd_study_instance_uid = self.series_instances_dictionary[index]['StudyInstanceUID']
+        RDs = self.series_instances_dictionary[index]['RDs']
+        reader = sitk.ImageFileReader()
+        output, spacing, direction, origin = None, None, None, None
+        self.dose = None
+        self.dose_handle = None
+        for rd_series_instance_uid in RDs:
+            rd = RDs[rd_series_instance_uid]
+            dose_file = rd['Path']
+            reader.SetFileName(dose_file)
+            reader.ReadImageInformation()
+            dose = reader.Execute()
+            spacing = dose.GetSpacing()
+            origin = dose.GetOrigin()
+            direction = dose.GetDirection()
+            scaling_factor = float(reader.GetMetaData("3004|000e"))
+            dose = sitk.GetArrayFromImage(dose) * scaling_factor
+            if output is None:
+                output = dose
+            else:
+                output += dose
+        if output is not None:
+            self.dose = output
+            output = sitk.GetImageFromArray(output)
+            output.SetSpacing(spacing)
+            output.SetDirection(direction)
+            output.SetOrigin(origin)
+            self.dose_handle = output
+
     def get_mask(self):
         assert self.index in self.series_instances_dictionary, \
             'Index is not present in the dictionary! Set it using set_index(index)'
         assert self.Contour_Names, 'If you want a mask, you need to set the contour names you are looking ' \
-                                   'for, use set_contour_names(list_of_roi_names).\nIf you just want to look at images ' \
-                                   'use get_images() not get_images_and_mask() or get_mask()'
+                                   'for, use set_contour_names_and_associations(list_of_roi_names).\nIf you just' \
+                                   ' want to look at images  use get_images() not get_images_and_mask() or get_mask()'
         index = self.index
         if self.dicom_handle_uid != self.series_instances_dictionary[index]['SeriesInstanceUID']:
             print('Loading images for index {}, since mask was requested but image loading was '
@@ -623,16 +736,9 @@ class DicomReaderWriter:
         if pixel_id.find('int') == -1:
             self.annotation_handle = sitk.Cast(self.annotation_handle, sitk.sitkUInt8)
         sitk.WriteImage(self.annotation_handle, annotation_path)
-        if len(self.dose_handles) > 0:
-            for dose_index, dose_handle in enumerate(self.dose_handles):
-                if len(self.dose_handles) > 1:
-                    dose_path = os.path.join(out_path,
-                                             'Overall_dose_{}_{}_{}.nii.gz'.format(self.desciption, self.iteration,
-                                                                                   dose_index))
-                else:
-                    dose_path = os.path.join(out_path,
-                                             'Overall_dose_{}_{}.nii.gz'.format(self.desciption, self.iteration))
-                sitk.WriteImage(dose_handle, dose_path)
+        if self.dose_handle is not None:
+            dose_path = os.path.join(out_path, 'Overall_dose_{}_{}.nii.gz'.format(self.desciption, self.iteration))
+            sitk.WriteImage(self.dose_handle, dose_path)
         fid = open(os.path.join(self.series_instances_dictionary[self.index]['Image_Path'],
                                 '{}_Iteration_{}.txt'.format(self.desciption, self.iteration)), 'w+')
         fid.close()
@@ -642,7 +748,6 @@ class DicomReaderWriter:
         :param prediction_array: numpy array of prediction, expected shape is [#Images, Rows, Cols, #Classes + 1]
         :param output_dir: directory to pass RT structure to
         :param ROI_Names: list of ROI names equal to the number of classes
-        :param index: index of the series instance UID to match with
         :return:
         """
         assert ROI_Names is not None, 'You need to provide ROI_Names'
@@ -814,7 +919,7 @@ class DicomReaderWriter:
         if os.path.exists(out_name):
             out_name = os.path.join(self.output_dir,
                                     'RS_MRN' + self.RS_struct.PatientID + '_' + self.RS_struct.SeriesInstanceUID + '1.dcm')
-        print('Writing out data...')
+        print('Writing out data...{}'.format(self.output_dir))
         pydicom.write_file(out_name, self.RS_struct)
         fid = open(os.path.join(self.output_dir, 'Completed.txt'), 'w+')
         fid.close()
@@ -862,37 +967,13 @@ class DicomReaderWriter:
                 continue
         return None
 
-    def get_dose(self):
-        reader = sitk.ImageFileReader()
-        output, spacing, direction, origin = None, None, None, None
-        for dose_file in self.RDs_in_case:
-            if os.path.split(dose_file)[-1].startswith('RTDOSE - PLAN'):
-                reader.SetFileName(dose_file)
-                reader.ReadImageInformation()
-                dose = reader.Execute()
-                spacing = dose.GetSpacing()
-                origin = dose.GetOrigin()
-                direction = dose.GetDirection()
-                scaling_factor = float(reader.GetMetaData("3004|000e"))
-                dose = sitk.GetArrayFromImage(dose) * scaling_factor
-                if output is None:
-                    output = dose
-                else:
-                    output += dose
-        if output is not None:
-            output = sitk.GetImageFromArray(output)
-            output.SetSpacing(spacing)
-            output.SetDirection(direction)
-            output.SetOrigin(origin)
-            self.dose_handles.append(output)
-
     def Make_Contour_From_directory(self, PathDicom):
-        print('Please move over to using add_dicom_to_dictionary_from_path')
-        self.add_dicom_to_dictionary_from_path(PathDicom=PathDicom)
+        print('Please move over to using walk_through_folders() instead of Make_Contour_From_directory()')
+        self.walk_through_folders(input_path=PathDicom)
 
     def make_contour_from_directory(self, dicom_path):
-        print('Please move over to using add_dicom_to_dictionary_from_path')
-        self.add_dicom_to_dictionary_from_path(PathDicom=dicom_path)
+        print('Please move over to using walk_through_folders() instead of Make_Contour_From_directory()')
+        self.walk_through_folders(input_path=dicom_path)
         return None
 
     def rewrite_RT(self, lstRSFile=None):
